#include <config.h>			/* private file */

/* Copyright (c) 1979 Regents of the University of California */

#define MAXHOP	32	/* max number of tc= indirections */

#if defined(HAVE_CTYPE_H)
#include <ctype.h>
#endif

#if defined(HAVE_STDIO_H)
#include <stdio.h>
#endif

#if defined(HAVE_STRING_H)
#include <string.h>
#endif

#if defined(HAVE_STDLIB_H)
#include <stdlib.h>
#endif

#if defined(HAVE_FCNTL_H)
#include <fcntl.h>	/* for open() prototype */
#endif

#if defined(HAVE_UNISTD_H)
#include <unistd.h>	/* for close(), read(), write() prototypes */
#else
#if defined(HAVE_LIBC_H)/* NeXT systems: conflicts with unistd.h, so want only one of them */
#include <libc.h>	/* for close(), read(), write() prototypes */
#endif
#endif

#include <stdc.h>			/* private file */
#include <bufsiz.h>			/* private file */
#include <vgrindefs.h>			/* private file */

/*
 * grindcap - routines for dealing with the language definitions data base
 *	(code stolen almost totally from termcap)
 *
 * BUG:		Should use a "last" pointer in tbuf, so that searching
 *		for capabilities alphabetically would not be a n**2/2
 *		process when large numbers of capabilities are given.
 * Note:	If we add a last pointer now we will screw up the
 *		tc capability. We really should compile termcap.
 *
 * Essentially all the work here is scanning and decoding escapes
 * in string capabilities.  We don't use stdio because the editor
 * doesn't, and because living w/o it is not hard.
 */

static char		*tbuf;
static const char	*filename;
static int		hopcount; /* detect infinite loops in termcap, init 0 */

static const char	*tdecode ARGS((register const char *str_,
				       char **area_));
static int		tnamatch ARGS((const char *np_));
static const char	*tskip ARGS((register const char *bp_));

/*
 * Get an entry for terminal name in buffer bp,
 * from the termcap file.  Parse is very rudimentary;
 * we just notice escaped newlines.
 */
#if STDC
int
tgetent(char *bp, const char *name, const char *file)
#else
int
tgetent(bp, name, file)
	char *bp;
	const char *name;
	const char *file;
#endif
{
	register char *cp;
	register int c;
	register int i = 0, cnt = 0;
	char ibuf[MAXBUF];
	int tf;

	tbuf = bp;
	tf = 0;
	filename = file;
	tf = open(filename, 0);
	if (tf < 0)
		return (-1);
	for (;;) {
		cp = bp;
		for (;;) {
			if (i == cnt) {
				cnt = read(tf, ibuf, MAXBUF);
				if (cnt <= 0) {
					close(tf);
					return (0);
				}
				i = 0;
			}
			c = ibuf[i++];
			if (c == '\n') {
				if (cp > bp && cp[-1] == '\\'){
					cp--;
					continue;
				}
				break;
			}
			if (cp >= bp+MAXBUF) {
				write(2,"Vgrind entry too long\n", 23);
				break;
			} else
				*cp++ = c;
		}
		*cp = 0;

		/*
		 * The real work for the match.
		 */
		if (tnamatch(name)) {
			close(tf);
			return(tnchktc());
		}
	}
}

/*
 * tnchktc: check the last entry, see if it's tc=xxx. If so,
 * recursively find xxx and append that entry (minus the names)
 * to take the place of the tc=xxx entry. This allows termcap
 * entries to say "like an HP2621 but doesn't turn on the labels".
 * Note that this works because of the left to right scan.
 */
int
tnchktc(VOID_ARG)
{
	register char *p, *q;
	char tcname[16];	/* name of similar terminal */
	char tcbuf[MAXBUF];
	char *holdtbuf = tbuf;
	int l;

	p = tbuf + strlen(tbuf) - 2;	/* before the last colon */
	while (*--p != ':')
		if (p<tbuf) {
			write(2, "Bad vgrind entry\n", 18);
			return (0);
		}
	p++;
	/* p now points to beginning of last field */
	if (p[0] != 't' || p[1] != 'c')
		return(1);
	strcpy(tcname,p+3);
	q = tcname;
	while (q && *q != ':')
		q++;
	*q = 0;
	if (++hopcount > MAXHOP) {
		write(2, "Infinite tc= loop\n", 18);
		return (0);
	}
	if (tgetent(tcbuf, tcname, filename) != 1)
		return(0);
	for (q=tcbuf; *q != ':'; q++)
		;
	l = p - holdtbuf + strlen(q);
	if (l > MAXBUF) {
		write(2, "Vgrind entry too long\n", 23);
		q[MAXBUF - (p-tbuf)] = 0;
	}
	strcpy(p, q+1);
	tbuf = holdtbuf;
	return(1);
}

/*
 * Tnamatch deals with name matching.  The first field of the termcap
 * entry is a sequence of names separated by |'s, so we compare
 * against each such name.  The normal : terminator after the last
 * name (before the first field) stops us.
 */
#if STDC
static int
tnamatch(const char *np)
#else
static int
tnamatch(np)
	const char *np;
#endif
{
	register const char *Np, *Bp;

	Bp = tbuf;
	if (*Bp == '#')
		return(0);
	for (;;) {

#define TOLOWER(c) (isupper(c) ? tolower(c) : (c))

		for (Np = np; *Np && TOLOWER(*Bp) == TOLOWER(*Np); Bp++, Np++)
			continue;

#undef TOLOWER

		if (*Np == 0 && (*Bp == '|' || *Bp == ':' || *Bp == 0))
			return (1);
		while (*Bp && *Bp != ':' && *Bp != '|')
			Bp++;
		if (*Bp == 0 || *Bp == ':')
			return (0);
		Bp++;
	}
}

/*
 * Skip to the next field.  Notice that this is very dumb, not
 * knowing about \: escapes or any such.  If necessary, :'s can be put
 * into the termcap file in octal.
 */
#if STDC
static const char *
tskip(register const char *bp)
#else
static const char *
tskip(bp)
	register const char *bp;
#endif
{

	while (*bp && *bp != ':')
		bp++;
	if (*bp == ':')
		bp++;
	return (bp);
}

/*
 * Return the (numeric) option id.
 * Numeric options look like
 *	li#80
 * i.e. the option string is separated from the numeric value by
 * a # character.  If the option is not found we return -1.
 * Note that we handle octal numbers beginning with 0.
 */
#if STDC
int
tgetnum(const char *id)
#else
int
tgetnum(id)
	const char *id;
#endif
{
	register int i, base;
	register const char *bp = tbuf;

	for (;;) {
		bp = tskip(bp);
		if (*bp == 0)
			return (-1);
		if (*bp++ != id[0] || *bp == 0 || *bp++ != id[1])
			continue;
		if (*bp == '@')
			return(-1);
		if (*bp != '#')
			continue;
		bp++;
		base = 10;
		if (*bp == '0')
			base = 8;
		i = 0;
		while (isdigit(*bp))
			i *= base, i += *bp++ - '0';
		return (i);
	}
}

/*
 * Handle a flag option.
 * Flag options are given "naked", i.e. followed by a : or the end
 * of the buffer.  Return 1 if we find the option, or 0 if it is
 * not given.
 */
#if STDC
int
tgetflag(const char *id)
#else
int
tgetflag(id)
	const char *id;
#endif
{
	register const char *bp = tbuf;

	for (;;) {
		bp = tskip(bp);
		if (!*bp)
			return (0);
		if (*bp++ == id[0] && *bp != 0 && *bp++ == id[1]) {
			if (!*bp || *bp == ':')
				return (1);
			else if (*bp == '@')
				return(0);
		}
	}
}

/*
 * Get a string valued option.
 * These are given as
 *	cl=^Z
 * Much decoding is done on the strings, and the strings are
 * placed in area, which is a ref parameter which is updated.
 * No checking on area overflow.
 */
#if STDC
const char *
tgetstr(const char *id, char **area)
#else
const char *
tgetstr(id, area)
	const char *id;
	char **area;
#endif
{
	register const char *bp = tbuf;

	for (;;) {
		bp = tskip(bp);
		if (!*bp)
			return (0);
		if (*bp++ != id[0] || *bp == 0 || *bp++ != id[1])
			continue;
		if (*bp == '@')
			return(0);
		if (*bp != '=')
			continue;
		bp++;
		return (tdecode(bp, area));
	}
}

/*
 * Tdecode does the grung work to decode the
 * string capability escapes.
 */
#if STDC
static const char *
tdecode(register const char *str, char **area)
#else
static const char *
tdecode(str, area)
	register const char *str;
	char **area;
#endif
{
	register char *cp;
	register int c;

	cp = *area;
	while ((c = *str++, c)) {
	    if (c == ':' && *(cp-1) != '\\')
		break;
#if 1
	    /* The old code provided no way for a single backslash to be */
	    /* represented in a string.  The new code allows one to use */
	    /* :xx=\\:. */
	    if ((c == ':') && (*(cp-1) == '\\') && (*(cp-2) == '\\'))
	    {
		cp--;			/* backup one character */
		break;
	    }
#endif
	    *cp++ = c;
	}
	*cp++ = 0;
	str = *area;
	*area = cp;
	return (str);
}
